/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * arch-os.c: architecture/os dependent functions.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include "threadP.h"
#include <stdio.h>

typedef long unsigned int Address; /* see "common/h/Types.h" */

/* put this around an ifdef -- linux only! */
#if defined(i386_unknown_linux2_0)
    #undef setjmp
    #define setjmp __setjmp
#endif
/* this shouldn't have been needed, since setjmp.h does this, but
for some reason the macro was undefined/redefined to be 'setjmp' */





/************************************************************************
 * types.
************************************************************************/

typedef enum {
    DIRECTION_DOWNWARD = -1,
    DIRECTION_UNKNOWN  = 0,
    DIRECTION_UPWARD   = 1
}   Direction;




/************************************************************************
 * symbolic constants.
************************************************************************/

/* define ARCH_STACK_DIRECTION in the Makefile of type Direction */
static const Direction STACK_DIRECTION = ARCH_STACK_DIRECTION;
static const unsigned  MIN_STACK_SLACK = 4096;




/************************************************************************
 * global variables.
************************************************************************/

static unsigned  stack_align     = 0;





/************************************************************************
 * unsigned round(unsigned val, unsigned align, Direction dir)
 *
 * round `val' to an alignment of `align' either "down" or "up"
 * depending on the direction `dir'.
************************************************************************/

static unsigned
round(unsigned val, unsigned align, Direction dir)
{
    unsigned diff;

    diff = val % align;
    if (dir == DIRECTION_DOWNWARD) {
        val -= diff;
    }
    else if (diff > 0) {
        val += (align - diff);
    }

    return val;
}





/************************************************************************
 * void thr_adjust_stack(Thread t)
 *
 * adjust `t->exec.stack' to point to the proper "end" based on
 * `t->exec.stack_size'.
************************************************************************/

void
thr_adjust_stack(Thread t)
{
    Address bot, top;

    assert(stack_align != 0);

    bot = (Address) t->exec.stack;
    top = (Address) t->exec.stack + t->exec.stack_size;

    bot = round(bot, stack_align, DIRECTION_UPWARD);
    top = round(top, stack_align, DIRECTION_DOWNWARD);

#if (defined(ALLOCATE_SCRATCH_WINDOW) && defined(SCRATCH_WINDOW_SIZE))
    bot += SCRATCH_WINDOW_SIZE;
#endif /* (defined(ALLOCATE_SCRATCH_WINDOW) && defined(SCRATCH_WINDOW_SIZE)) */

    if (STACK_DIRECTION == DIRECTION_UPWARD) {
        t->exec.st_bot = (void *) bot;
        t->exec.st_top = (void *) top;
    }
    else {
        t->exec.st_bot = (void *) top;
        t->exec.st_top = (void *) bot;
    }

    return;
}



#if defined(rs6000_ibm_aix3_2) || defined(rs6000_ibm_aix4_1)
void rs6000restore(jmp_buf context, int flag)
{
    /* load the stack pointer to make the restore routine happy */
    asm("l 	1, 12(3)");
    longjmp(context, flag);
}
#endif


/************************************************************************
 * void thr_init_arch_magic(void)
 *
 * initialize architecture specific functions and variables. currently,
 * this is the Tcb::save_context() and Tcb::restore_context()
 * functions.
 *
 * since the save and restore functions can be different if preemption
 * is desired in libthread, two versions of this function are defined.
************************************************************************/

void
thr_init_arch_magic(void)
{
#undef FOUND_SUITABLE_ARCH_MAGIC



#if defined(rs6000_ibm_aix3_2) || defined(rs6000_ibm_aix4_1)
#define FOUND_SUITABLE_ARCH_MAGIC 1
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (rs6000restore);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(rs6000_ibm_aix3_2) */


#if defined(i386_unknown_bsd)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(i386_unknown_bsd) */


#if defined(hppa1_1_hp_hpux)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(hppa1_1_hp_hpux) */


#if defined(mips_dec_ultrix4_3)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (Rest_Func) ((int) (siglongjmp) + 20);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (Rest_Func) ((int) (longjmp) + 20);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 16;
#endif /* defined(mips_dec_ultrix4_3) */


#if defined(i386_unknown_solaris2_5)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(i386_unknown_solaris2_5) */


#if defined(i386_unknown_linux2_0)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    /* Under Linux glibc, we need to use a macro, see
     * THR_SAVE_CONTEXT in threadP.h for an explanation */
    /* Tcb::save_context    = (setjmp); */
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(i386_unknown_linux2_0) */


#if defined(sparc_sun_solaris2_4)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 8;
#endif /* defined(sparc_sun_sunos4_1_3) */

#if defined(mips_sgi_irix6_4)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
#error preemption for IRIX platform not done yet
    Tcb::save_context    = (sigsetjmp);
    Tcb::restore_context = (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
    Tcb::save_context    = (setjmp);
    Tcb::restore_context = (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
    stack_align         = 16;
#endif /* defined(mips_sgi_irix6_4) */



#if defined(i386_unknown_nt4_0)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
#error preemption not currently supported under Windows NT due to lack of context switching support
	Tcb::save_context	= (sigsetjmp);
	Tcb::restore_context	= (siglongjmp);
#else /* defined(ENABLE_PREEMPTION) */
	Tcb::save_context	= (setjmp);
	Tcb::restore_context	= (longjmp);
#endif /* defined(ENABLE_PREEMPTION) */
	stack_align			= 8;
#endif /* defined(i386_unknown_nt4_0) */


#if !defined(FOUND_SUITABLE_ARCH_MAGIC)
#error "no suitable architecture magic found. check compilation flags"
#endif /* !defined(FOUND_SUITABLE_ARCH_MAGIC) */

    assert(stack_align != 0);
    assert(STACK_DIRECTION != DIRECTION_UNKNOWN);

    return;
}





/************************************************************************
 * void thr_do_arch_magic(Thread t)
 *
 * save the current environment into `t->exec.state', reset stack
 * pointers and save the address of the function `Tcb::dummy_start' in
 * the pc of the environment. also round `t->exec.stack' to the proper
 * alignment.
 *
 * the entries that need to be saved and their save locations can be
 * different if preemption is desired in libthread.
************************************************************************/

void
thr_do_arch_magic(Thread t)
{
#undef FOUND_SUITABLE_ARCH_MAGIC

#if defined(i386_unknown_bsd)
#define FOUND_SUITABLE_ARCH_MAGIC
/************************************************************************
 * state[0] = saved return address
 * state[2] = saved stack pointer
 * state[3] = saved frame pointer
 *
 * stack are aligned on 8-byte boundaries, just to be safe.
 * layout of the environment buffer is identical for preemptive
 * and non-preemptive threads.
************************************************************************/

    t->exec.state[0] = (int) Tcb::dummy_start;
    t->exec.state[2] = (int) t->exec.st_bot;
    t->exec.state[3] = (int) t->exec.st_bot;

#endif /* defined(i386_unknown_bsd) */


#if defined(i386_unknown_solaris2_5)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)

#error preemption for X86 Solaris platform not done yet

#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state[5] = saved return address
 * state[4] = saved stack pointer
 * state[3] = saved frame pointer
 *
 * stack are aligned on 8-byte boundaries.
************************************************************************/

    t->exec.state[5] = (int) Tcb::dummy_start;
    t->exec.state[4] = (int) t->exec.st_bot;
    t->exec.state[3] = (int) t->exec.st_bot;

#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(i386_unknown_solaris2_5) */


#if defined(i386_unknown_linux2_0)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)

#error preemption for X86 Linux platform not done yet

#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state->__sp = saved stack pointer
 * state->__pc = saved pc
 *
 * stack are aligned on 8-byte boundaries.
************************************************************************/

    t->exec.state[0].__jmpbuf[JB_SP] = (int) t->exec.st_bot;
    t->exec.state[0].__jmpbuf[JB_BP] = (int) t->exec.st_bot;
    t->exec.state[0].__jmpbuf[JB_PC] = (int)((void*) Tcb::dummy_start);

#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(i386_unknown_linux2_0) */


#if defined(rs6000_ibm_aix3_2) || defined(rs6000_ibm_aix4_1)
/************************************************************************
 * state[2] = saved return address
 * state[3] = saved stack pointer ($sp == $r1)
 * On POWER, the address of the function is an indirect handle.
 *
 * stack are aligned on 8-byte boundaries, just to be safe.
 * layout of the environment buffer is identical for preemptive
 * and non-preemptive threads.
************************************************************************/

    t->exec.state[2] = (int) *((int*)Tcb::dummy_start);
    t->exec.state[3] = (int) t->exec.st_bot;
    t->exec.state[23] = (int) t->exec.st_bot;
    return;

#endif /* defined(rs6000_ibm_aix3_2) */

#if defined(hppa1_1_hp_hpux)
#define FOUND_SUITABLE_ARCH_MAGIC
/************************************************************************
 * state[0] = saved return address + saved stack pointer
 *
 * stack are aligned on 8-byte boundaries, just to be safe.
 * layout of the environment buffer is identical for preemptive
 * and non-preemptive threads.
 *
 * the programs using libthread _MUST_ be linked static. the layout
 * of the environment buffer is different for static and shared
 * binaries.
************************************************************************/

    {
    int* ip = (int *) &(t->exec.state[0]);
    int  pc = (int) Tcb::dummy_start;
    int  sp = (int) t->exec.st_bot;

    memcpy(&ip[0], &pc, sizeof ip);
    memcpy(&ip[1], &sp, sizeof ip);
    }

#endif /* defined(hppa1_1_hp_hpux) */


#if defined(mips_dec_ultrix4_3)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
/************************************************************************
 * state[2]  = saved return address
 * state[25] = saved start of stack
 * state[32] = saved end of stack
 *
 * stack are aligned on 16-byte boundaries.
 * longjmp problem exists on ultrix.
************************************************************************/

    t->exec.state[2]  = (int) Tcb::dummy_start;
    t->exec.state[25] = (int) t->exec.st_top;
    t->exec.state[32] = (int) t->exec.st_bot;

#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state[2]  = saved return address
 * state[25] = saved start of stack
 * state[32] = saved end of stack
 *
 * stack are aligned on 16-byte boundaries.
 * longjmp problem exists on ultrix.
************************************************************************/

    t->exec.state[2]  = (int) Tcb::dummy_start;
    t->exec.state[25] = (int) t->exec.st_top;
    t->exec.state[32] = (int) t->exec.st_bot;

#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(mips_dec_ultrix4_3) */


#if defined(sparc_sun_solaris2_4)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
/************************************************************************
 * state[1] = saved stack pointer
 * state[2] = saved pc
 *
 * stack are aligned on 8-byte boundaries.
 * also need a trap somewhere here to flush register windows, not!
************************************************************************/

    t->exec.state[1] = (int) t->exec.st_bot;
    t->exec.state[2] = (int) Tcb::dummy_start;

#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state[1] = saved stack pointer
 * state[2] = saved pc
 *
 * stack are aligned on 8-byte boundaries.
 * also need a trap somewhere here to flush register windows, not!
************************************************************************/

    t->exec.state[1] = (int) t->exec.st_bot;
    t->exec.state[2] = (int) Tcb::dummy_start;

#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(sparc_sun_solaris2_4) */


#if defined(sparc_sun_sunos4_1_3)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
/************************************************************************
 * state[3] = saved stack pointer (%o6)
 * state[4] = saved pc            (%o7)
 * state[5] = saved npc
 *
 * stack are aligned on 8-byte boundaries.
 * also need a trap somewhere here to flush register windows, not!
************************************************************************/

    t->exec.state[3] = (int) t->exec.st_bot;
    t->exec.state[4] = (int) Tcb::dummy_start;
    t->exec.state[5] = (int) Tcb::dummy_start + 4;

#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state[2] = saved stack pointer (%o6)
 * state[3] = saved pc            (%o7)
 * state[4] = saved npc
 *
 * stack are aligned on 8-byte boundaries.
 * also need a trap somewhere here to flush register windows, not!
************************************************************************/

    t->exec.state[2] = (int) t->exec.st_bot;
    t->exec.state[3] = (int) Tcb::dummy_start;
    t->exec.state[4] = (int) Tcb::dummy_start + 4;

#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(sparc_sun_sunos4_1_3) */


#if defined(mips_sgi_irix6_4)
#define FOUND_SUITABLE_ARCH_MAGIC
#if defined(ENABLE_PREEMPTION)
#error preemption for IRIX platform not done yet
#else /* defined(ENABLE_PREEMPTION) */
/************************************************************************
 * state[JB_PC]  = return address
 * state[JB_SP]  = stack pointer
 *
 * JB_xx buffer offset macros from <setjmp.h>
 * stacks are aligned on 16-byte boundaries - MIPSpro N32 ABI Handbook
************************************************************************/
    t->exec.state[JB_PC]  = (__uint64_t) Tcb::dummy_start;
    t->exec.state[JB_SP]  = (__uint64_t) t->exec.st_bot;
#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(mips_sgi_irix6_4) */


#if defined(i386_unknown_nt4_0)
#if defined(ENABLE_PREEMPTION)
#error Preemption not yet available for i386 NT
#else /* defined(ENABLE_PREEMPTION) */
#define FOUND_SUITABLE_ARCH_MAGIC
/************************************************************************
 * _JUMP_BUFFER.Esp = saved stack pointer
 * _JUMP_BUFFER.Ebp = saved frame pointer
 * _JUMP_BUFFER.Eip = saved instruction pointer
************************************************************************/
	{
		_JUMP_BUFFER* pjb = (_JUMP_BUFFER*)(t->exec.state);

		pjb->Esp = (unsigned long)t->exec.st_bot - 4;
		pjb->Ebp = (unsigned long)t->exec.st_bot;
		pjb->Eip = (unsigned long)Tcb::dummy_start;
	}
#endif /* defined(ENABLE_PREEMPTION) */
#endif /* defined(i386_unknown_nt4_0 */


#define FOUND_SUITABLE_ARCH_MAGIC
#if !defined(FOUND_SUITABLE_ARCH_MAGIC)
#error "no suitable architecture magic found. check compilation flags"
#endif /* !defined(FOUND_SUITABLE_ARCH_MAGIC) */

    return;
}





/************************************************************************
 * void thr_check_stack(void)
 *
 * check to see if the current thread is approaching its stack boundary
 * too closely and print a warning.
************************************************************************/

void
thr_check_stack(void)
{
    unsigned slack;

    assert(thr_curr);
    if (thr_curr->flags.is_main) {
        return;
    }

    if (STACK_DIRECTION == DIRECTION_UPWARD) {
        assert((Address) thr_curr->exec.st_top >= (Address) &slack);
        assert((Address) thr_curr->exec.st_bot <= (Address) &slack);
        slack = ((Address) thr_curr->exec.st_top - (Address) &slack);
    }
    else {
        assert((Address) &slack >= (Address) thr_curr->exec.st_top);
        assert((Address) &slack <= (Address) thr_curr->exec.st_bot);
        slack = ((Address) &slack - (Address) thr_curr->exec.st_top);
    }

    DO_TRACE_1(("check_stack(): slack is %u", slack));

    if (slack < MIN_STACK_SLACK) {
        (void) fprintf(stderr, "*** thread [%u,%s] stack slack is %u ***\n",
            thr_curr->id.tid, thr_curr->id.name, slack);
    }

    return;
}
